/*****************************************************************************

                'J' - Yet Another Molecular Mechanics Program

                           (c) 1995 Jan T Pedersen


This program is not in the public domain and may not be used or distributed
by anyone without the written permission of the author. The functionality
of the whole or parts of the program is not guaranteed.

******************************************************************************
 $Header$
 $Revision$
*****************************************************************************/

#include "Graph.h"
#include "GraphStatics.h"

/*
 * This subroutine finds the maximal complete subgraphs (cliques) of a
 * given undirected graph. The algorithm consists in a recursively defined
 * extension operator (extend) applied to the three following sets:
 * set compsub: set to be extended by a new point or shrunk by one point as
 *              the algorithm travels along a branch of the backtracking tree.
 * set cand:    set of all vertices connected to the current vertex in compsub
 *              i.e., vertices which will be used as an extension of the
 *              present configuration of compsub.
 * set not:     set of all points which have already served as an extension and
 *              are now excluded.
 *
 * The basic mechanism consists in the following steps:
 * Step 1:  selection of a candidate.
 * Step 2:  creating new sets cand and not from the old sets by removing all points
 *          not connected to the selected candidate, keeping the old sets intact.
 * Step 3:  adding the selected candidate to compsub.
 * Step 4:  calling the extension operator to operate on the sets just formed.
 * Step 5:  upon return the selected candidate is transfered from compsub to the
 *          old set not.
 *
 * The set compsub is a stack external to the subroutine.
 * The sets not and cand are stored in the local array old,
 * not from positions 1 to ne and cand from positions ne+1 to ce.
 * nvert x nvert array [connected] indicates which vertices are connected to which
 * vertices. This array is external to the subroutine.
 *
 * Last modification: 09/09/99
 *
 * Reference: Bron C. and Kerbosch J., Commun. A.C.M., vol. 16, pp 575-577, 1973.
 *
 */

int Extend(int old[], int ne, int ce)

{
int new[MAX_NODES];
int knod, nod, fixp;
int newne, newce, i, j, count, pos, p, s, sel, minnod;

#ifdef DBG
fprintf (stdout,"DBG >> Extend()\n");
#endif

/* NSC: initialize to zero: */

knod=0;
nod=0;
fixp=0;
newne=0;
newce=0;
i=0;
j=0;
count=0;
pos=0;
p=0;
s=0;
sel=0;
minnod=0;

/* the algorithm: */
minnod = ce;
nod = 0;

/*
 * Look for the vertex that is connected to the maximum number of vertices, i.e.,
 * the minimum number of disconnections (Step 1)
 *
 */
for(i = 1; (i <= ce) && (minnod != 0); i++)
	{
	p = old[i];
	count = 0;

	/* Count disconnections */
	for(j = ne + 1; (j <= ce) && (count < minnod); j++)
		{
		if(!connected[p][old[j]])
			{
			count++;
			pos = j;	/* save position of potential candidate */
			}
		}
/* 
 * Test whether vertex at position i in the 
 * stack has a minimum number of diconnections 
 */
	if(count < minnod)
		{
		fixp = p;
		minnod = count;
		if(i <= ne)
			s = pos;
		else
			{
			s = i;
			nod = 1;	/* Number of disconnections preincreased by one */
			}
		}
	}

/*
 * Backtrack Cycle: the loop is executed as many 
 * times as there are vertices not connected to the selected point
 */

/* !!! is >= correct ?!!! */
for(knod = minnod + nod; knod >= 1; knod--)
/* for(knod = minnod + nod; knod > 1; knod--) */
	{
/*
 * Interchange
 */
	p = old[s];
	old[s] = old[ne+1];
	old[ne+1] = p;
	sel = old[ne+1];
/*
 * Fill new set not (Step 2)
 */
	newne = 0;
	for(i = 1; i <= ne; i++)
		if(connected[sel][old[i]])
			new[++newne] = old[i];
/*
 * Fill new set cand (Step 2)
 */
	newce = newne;
	for(i = ne + 2; i <= ce; i++)
		if(connected[sel][old[i]])
			new[++newce] = old[i];

/*
 * The selected vertex is added to compsub (Step 3)
 */
	compsub[++nel] = sel; /* Note that index nel is external to the subroutine */

/*
 * Set cand and not are empty, a new clique has been found:
 * the recursive process terminates when this condition is met.
 * The new clique is not recorded if it has less than mincliq elements, or if
 * the value returned by clique_proba is negative (indicative of a poor rmsd).
 * The new clique is then compared with the cliques already stored. If is has
 * no element in common with these cliques it is stored. If it has at least
 * one element in common with a clique, and if its probability is greater# than
 * the probability of the clique already stored this later is replaced by the
 * new one otherwise the new clique is not recorded. Note that if the two cliques
 * differ by more than DELTA elements it's useless to calculate the probability
 * of the new clique which is very likely smaller.
 */

	if(newce == 0)
		{
/* 
 * The new clique is stored in compsub. 
 */
		if (nel>maxcliquesize)
			{
			maxcliquesize=nel;
/* 
 * store the max clique: 
 */
			for (i=1;i<=maxcliquesize;i++)
				maxclique[i]=compsub[i];
			}

/* 
 * clique weight not part of the algorithm 
 */
/*!!!			clique_weight(&cliques[clique_no]) = (float)0.0; */
/*!!!			for(i = 1; i <= nel; i++) */
/*!!!			{ */
/*!!!				clique_weight(&cliques[clique_no]) += (float)1.0; */
							/* clique_weight(&cliques[clique_no]) += (((int) compsub[i]) * 1.0); */
							/* clique_vertex(&cliques[clique_no], i) = compsub[i]; */
/*!!!			} */
/*!!!			clique_size(&cliques[clique_no]) = nel; */
/*!!!			clique_no++;	*/
/*!!!NSC:
		printf("%3d %7.2f | %3d | ", clique_no-1, clique_weight(&cliques[clique_no-1]), clique_size(&cliques[clique_no-1]));
		for(i = 1; i <= clique_size(&cliques[clique_no-1]); i++)
			printf("%d ", compsub[i]);
		printf("\n");
*/
		}
/*
 * Otherwise calls the extension operator to operate on the sets just formed	(Step 4)
 */
	else
		{
		if (newne < newce)
			Extend(new, newne, newce);
		}

/*
 * Upon return removal of the selected candidate from compsub and its addition to the old set not (Step 5)
 *
 * Remove from compsub */
	nel--;
/* Add to not */
	ne++;

/*
 * Select a candidate disconnected to the fixed point (Step 1)
 */
	if( knod > 1)
		{
		s = ne + 1;
		while(connected[fixp][old[s]])
			s++;
		}
	} /* End of the loop with index knod */
#ifdef DBG
fprintf (stdout,"DBG >> Extend() - done\n");
#endif
return (TRUE);
}

